package com.sun.org.apache.xml.internal.serializer;

import java.util.Set;
import com.sun.org.apache.xml.internal.serializer.dom3.DOM3SerializerImpl;
import org.xml.sax.ext.Locator2;
import org.xml.sax.SAXParseException;
import org.w3c.dom.Node;
import com.sun.org.apache.xml.internal.serializer.utils.Utils;
import java.io.IOException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import java.util.HashMap;
import org.xml.sax.Locator;
import java.io.Writer;
import javax.xml.transform.SourceLocator;
import javax.xml.transform.Transformer;

public abstract class SerializerBase implements SerializationHandler, SerializerConstants
{
    protected boolean m_needToCallStartDocument;
    protected boolean m_cdataTagOpen;
    protected AttributesImplSerializer m_attributes;
    protected boolean m_inEntityRef;
    protected boolean m_inExternalDTD;
    protected String m_doctypeSystem;
    protected String m_doctypePublic;
    boolean m_needToOutputDocTypeDecl;
    protected boolean m_shouldNotWriteXMLHeader;
    private String m_standalone;
    protected boolean m_standaloneWasSpecified;
    protected boolean m_isStandalone;
    protected boolean m_doIndent;
    protected int m_indentAmount;
    protected String m_version;
    protected String m_mediatype;
    private Transformer m_transformer;
    protected NamespaceMappings m_prefixMap;
    protected SerializerTrace m_tracer;
    protected SourceLocator m_sourceLocator;
    protected Writer m_writer;
    protected ElemContext m_elemContext;
    protected char[] m_charsBuff;
    protected char[] m_attrBuff;
    private Locator m_locator;
    protected boolean m_needToCallSetDocumentInfo;
    protected String m_StringOfCDATASections;
    boolean m_docIsEmpty;
    protected HashMap<String, HashMap<String, String>> m_CdataElems;
    private HashMap<String, String> m_OutputProps;
    private HashMap<String, String> m_OutputPropsDefault;
    
    public SerializerBase() {
        this.m_needToCallStartDocument = true;
        this.m_cdataTagOpen = false;
        this.m_attributes = new AttributesImplSerializer();
        this.m_inEntityRef = false;
        this.m_inExternalDTD = false;
        this.m_needToOutputDocTypeDecl = true;
        this.m_shouldNotWriteXMLHeader = false;
        this.m_standaloneWasSpecified = false;
        this.m_isStandalone = false;
        this.m_doIndent = false;
        this.m_indentAmount = 0;
        this.m_version = null;
        this.m_writer = null;
        this.m_elemContext = new ElemContext();
        this.m_charsBuff = new char[60];
        this.m_attrBuff = new char[30];
        this.m_locator = null;
        this.m_needToCallSetDocumentInfo = true;
        this.m_StringOfCDATASections = null;
        this.m_docIsEmpty = true;
        this.m_CdataElems = null;
    }
    
    protected void fireEndElem(String name) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(4, name, (Attributes)null);
        }
    }
    
    protected void fireCharEvent(char[] chars, int start, int length) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(5, chars, start, length);
        }
    }
    
    @Override
    public void comment(String data) throws SAXException {
        int length = data.length();
        if (length > this.m_charsBuff.length) {
            this.m_charsBuff = new char[length * 2 + 1];
        }
        data.getChars(0, length, this.m_charsBuff, 0);
        this.comment(this.m_charsBuff, 0, length);
    }
    
    protected String patchName(String qname) {
        int lastColon = qname.lastIndexOf(58);
        if (lastColon > 0) {
            int firstColon = qname.indexOf(58);
            String prefix = qname.substring(0, firstColon);
            String localName = qname.substring(lastColon + 1);
            String uri = this.m_prefixMap.lookupNamespace(prefix);
            if (uri != null && uri.length() == 0) {
                return localName;
            }
            if (firstColon != lastColon) {
                return prefix + 58 + localName;
            }
        }
        return qname;
    }
    
    protected static String getLocalName(String qname) {
        int col = qname.lastIndexOf(58);
        return (col > 0) ? qname.substring(col + 1) : qname;
    }
    
    @Override
    public void setDocumentLocator(Locator locator) {
        this.m_locator = locator;
    }
    
    @Override
    public void addAttribute(String uri, String localName, String rawName, String type, String value, boolean XSLAttribute) throws SAXException {
        if (this.m_elemContext.m_startTagOpen) {
            this.addAttributeAlways(uri, localName, rawName, type, value, XSLAttribute);
        }
    }
    
    public boolean addAttributeAlways(String uri, String localName, String rawName, String type, String value, boolean XSLAttribute) {
        int index;
        if (localName == null || uri == null || uri.length() == 0) {
            index = this.m_attributes.getIndex(rawName);
        }
        else {
            index = this.m_attributes.getIndex(uri, localName);
        }
        boolean was_added;
        if (index >= 0) {
            this.m_attributes.setValue(index, value);
            was_added = false;
        }
        else {
            this.m_attributes.addAttribute(uri, localName, rawName, type, value);
            was_added = true;
        }
        return was_added;
    }
    
    @Override
    public void addAttribute(String name, String value) {
        if (this.m_elemContext.m_startTagOpen) {
            String patchedName = this.patchName(name);
            String localName = getLocalName(patchedName);
            String uri = this.getNamespaceURI(patchedName, false);
            this.addAttributeAlways(uri, localName, patchedName, "CDATA", value, false);
        }
    }
    
    @Override
    public void addXSLAttribute(String name, String value, String uri) {
        if (this.m_elemContext.m_startTagOpen) {
            String patchedName = this.patchName(name);
            String localName = getLocalName(patchedName);
            this.addAttributeAlways(uri, localName, patchedName, "CDATA", value, true);
        }
    }
    
    @Override
    public void addAttributes(Attributes atts) throws SAXException {
        for (int nAtts = atts.getLength(), i = 0; i < nAtts; ++i) {
            String uri = atts.getURI(i);
            if (null == uri) {
                uri = "";
            }
            this.addAttributeAlways(uri, atts.getLocalName(i), atts.getQName(i), atts.getType(i), atts.getValue(i), false);
        }
    }
    
    @Override
    public ContentHandler asContentHandler() throws IOException {
        return this;
    }
    
    @Override
    public void endEntity(String name) throws SAXException {
        if (name.equals("[dtd]")) {
            this.m_inExternalDTD = false;
        }
        this.m_inEntityRef = false;
        if (this.m_tracer != null) {
            this.fireEndEntity(name);
        }
    }
    
    @Override
    public void close() {
    }
    
    protected void initCDATA() {
    }
    
    @Override
    public String getEncoding() {
        return this.getOutputProperty("encoding");
    }
    
    @Override
    public void setEncoding(String encoding) {
        this.setOutputProperty("encoding", encoding);
    }
    
    @Override
    public void setOmitXMLDeclaration(boolean b) {
        String val = b ? "yes" : "no";
        this.setOutputProperty("omit-xml-declaration", val);
    }
    
    @Override
    public boolean getOmitXMLDeclaration() {
        return this.m_shouldNotWriteXMLHeader;
    }
    
    @Override
    public String getDoctypePublic() {
        return this.m_doctypePublic;
    }
    
    @Override
    public void setDoctypePublic(String doctypePublic) {
        this.setOutputProperty("doctype-public", doctypePublic);
    }
    
    @Override
    public String getDoctypeSystem() {
        return this.m_doctypeSystem;
    }
    
    @Override
    public void setDoctypeSystem(String doctypeSystem) {
        this.setOutputProperty("doctype-system", doctypeSystem);
    }
    
    @Override
    public void setDoctype(String doctypeSystem, String doctypePublic) {
        this.setOutputProperty("doctype-system", doctypeSystem);
        this.setOutputProperty("doctype-public", doctypePublic);
    }
    
    @Override
    public void setStandalone(String standalone) {
        this.setOutputProperty("standalone", standalone);
    }
    
    protected void setStandaloneInternal(String standalone) {
        if ("yes".equals(standalone)) {
            this.m_standalone = "yes";
        }
        else {
            this.m_standalone = "no";
        }
    }
    
    @Override
    public String getStandalone() {
        return this.m_standalone;
    }
    
    @Override
    public boolean getIndent() {
        return this.m_doIndent;
    }
    
    @Override
    public String getMediaType() {
        return this.m_mediatype;
    }
    
    @Override
    public String getVersion() {
        return this.m_version;
    }
    
    @Override
    public void setVersion(String version) {
        this.setOutputProperty("version", version);
    }
    
    @Override
    public void setMediaType(String mediaType) {
        this.setOutputProperty("media-type", mediaType);
    }
    
    @Override
    public int getIndentAmount() {
        return this.m_indentAmount;
    }
    
    @Override
    public void setIndentAmount(int m_indentAmount) {
        this.m_indentAmount = m_indentAmount;
    }
    
    @Override
    public void setIndent(boolean doIndent) {
        String val = doIndent ? "yes" : "no";
        this.setOutputProperty("indent", val);
    }
    
    @Override
    public void setIsStandalone(boolean isStandalone) {
        this.m_isStandalone = isStandalone;
    }
    
    @Override
    public void namespaceAfterStartElement(String uri, String prefix) throws SAXException {
    }
    
    @Override
    public DOMSerializer asDOMSerializer() throws IOException {
        return this;
    }
    
    private static final boolean subPartMatch(String p, String t) {
        return p == t || (null != p && p.equals(t));
    }
    
    protected static final String getPrefixPart(String qname) {
        int col = qname.indexOf(58);
        return (col > 0) ? qname.substring(0, col) : null;
    }
    
    @Override
    public NamespaceMappings getNamespaceMappings() {
        return this.m_prefixMap;
    }
    
    @Override
    public String getPrefix(String namespaceURI) {
        String prefix = this.m_prefixMap.lookupPrefix(namespaceURI);
        return prefix;
    }
    
    @Override
    public String getNamespaceURI(String qname, boolean isElement) {
        String uri = "";
        int col = qname.lastIndexOf(58);
        String prefix = (col > 0) ? qname.substring(0, col) : "";
        if ((!"".equals(prefix) || isElement) && this.m_prefixMap != null) {
            uri = this.m_prefixMap.lookupNamespace(prefix);
            if (uri == null && !prefix.equals("xmlns")) {
                throw new RuntimeException(Utils.messages.createMessage("ER_NAMESPACE_PREFIX", new Object[] { qname.substring(0, col) }));
            }
        }
        return uri;
    }
    
    @Override
    public String getNamespaceURIFromPrefix(String prefix) {
        String uri = null;
        if (this.m_prefixMap != null) {
            uri = this.m_prefixMap.lookupNamespace(prefix);
            return uri;
        }
        return uri;
    }
    
    @Override
    public void entityReference(String name) throws SAXException {
        this.flushPending();
        this.startEntity(name);
        this.endEntity(name);
        if (this.m_tracer != null) {
            this.fireEntityReference(name);
        }
    }
    
    @Override
    public void setTransformer(Transformer t) {
        this.m_transformer = t;
        if (this.m_transformer instanceof SerializerTrace && ((SerializerTrace)this.m_transformer).hasTraceListeners()) {
            this.m_tracer = (SerializerTrace)this.m_transformer;
        }
        else {
            this.m_tracer = null;
        }
    }
    
    @Override
    public Transformer getTransformer() {
        return this.m_transformer;
    }
    
    @Override
    public void characters(Node node) throws SAXException {
        this.flushPending();
        String data = node.getNodeValue();
        if (data != null) {
            int length = data.length();
            if (length > this.m_charsBuff.length) {
                this.m_charsBuff = new char[length * 2 + 1];
            }
            data.getChars(0, length, this.m_charsBuff, 0);
            this.characters(this.m_charsBuff, 0, length);
        }
    }
    
    @Override
    public void error(SAXParseException exc) throws SAXException {
    }
    
    @Override
    public void fatalError(SAXParseException exc) throws SAXException {
        this.m_elemContext.m_startTagOpen = false;
    }
    
    @Override
    public void warning(SAXParseException exc) throws SAXException {
    }
    
    protected void fireStartEntity(String name) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(9, name);
        }
    }
    
    private void flushMyWriter() {
        if (this.m_writer != null) {
            try {
                this.m_writer.flush();
            }
            catch (IOException ex) {}
        }
    }
    
    protected void fireCDATAEvent(char[] chars, int start, int length) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(10, chars, start, length);
        }
    }
    
    protected void fireCommentEvent(char[] chars, int start, int length) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(8, new String(chars, start, length));
        }
    }
    
    public void fireEndEntity(String name) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
        }
    }
    
    protected void fireStartDoc() throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(1);
        }
    }
    
    protected void fireEndDoc() throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(2);
        }
    }
    
    protected void fireStartElem(String elemName) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(3, elemName, this.m_attributes);
        }
    }
    
    protected void fireEscapingEvent(String name, String data) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(7, name, data);
        }
    }
    
    protected void fireEntityReference(String name) throws SAXException {
        if (this.m_tracer != null) {
            this.flushMyWriter();
            this.m_tracer.fireGenerateEvent(9, name, (Attributes)null);
        }
    }
    
    @Override
    public void startDocument() throws SAXException {
        this.startDocumentInternal();
        this.m_needToCallStartDocument = false;
    }
    
    protected void startDocumentInternal() throws SAXException {
        if (this.m_tracer != null) {
            this.fireStartDoc();
        }
    }
    
    protected void setDocumentInfo() {
        if (this.m_locator == null) {
            return;
        }
        try {
            String strVersion = ((Locator2)this.m_locator).getXMLVersion();
            if (strVersion != null) {
                this.setVersion(strVersion);
            }
        }
        catch (ClassCastException ex) {}
        return;
    }
    
    @Override
    public void setSourceLocator(SourceLocator locator) {
        this.m_sourceLocator = locator;
    }
    
    @Override
    public void setNamespaceMappings(NamespaceMappings mappings) {
        this.m_prefixMap = mappings;
    }
    
    @Override
    public boolean reset() {
        this.resetSerializerBase();
        return true;
    }
    
    private void resetSerializerBase() {
        this.m_attributes.clear();
        this.m_StringOfCDATASections = null;
        this.m_elemContext = new ElemContext();
        this.m_doctypePublic = null;
        this.m_doctypeSystem = null;
        this.m_doIndent = false;
        this.m_indentAmount = 0;
        this.m_inEntityRef = false;
        this.m_inExternalDTD = false;
        this.m_mediatype = null;
        this.m_needToCallStartDocument = true;
        this.m_needToOutputDocTypeDecl = false;
        if (this.m_prefixMap != null) {
            this.m_prefixMap.reset();
        }
        this.m_shouldNotWriteXMLHeader = false;
        this.m_sourceLocator = null;
        this.m_standalone = null;
        this.m_standaloneWasSpecified = false;
        this.m_tracer = null;
        this.m_transformer = null;
        this.m_version = null;
    }
    
    final boolean inTemporaryOutputState() {
        return this.getEncoding() == null;
    }
    
    @Override
    public void addAttribute(String uri, String localName, String rawName, String type, String value) throws SAXException {
        if (this.m_elemContext.m_startTagOpen) {
            this.addAttributeAlways(uri, localName, rawName, type, value, false);
        }
    }
    
    @Override
    public void notationDecl(String arg0, String arg1, String arg2) throws SAXException {
    }
    
    @Override
    public void unparsedEntityDecl(String arg0, String arg1, String arg2, String arg3) throws SAXException {
    }
    
    @Override
    public void setDTDEntityExpansion(boolean expand) {
    }
    
    void initCdataElems(String s) {
        if (s != null) {
            int max = s.length();
            boolean inCurly = false;
            boolean foundURI = false;
            StringBuilder buf = new StringBuilder();
            String uri = null;
            String localName = null;
            for (int i = 0; i < max; ++i) {
                char c = s.charAt(i);
                if (Character.isWhitespace(c)) {
                    if (!inCurly) {
                        if (buf.length() > 0) {
                            localName = buf.toString();
                            if (!foundURI) {
                                uri = "";
                            }
                            this.addCDATAElement(uri, localName);
                            buf.setLength(0);
                            foundURI = false;
                        }
                    }
                    else {
                        buf.append(c);
                    }
                }
                else if ('{' == c) {
                    inCurly = true;
                }
                else if ('}' == c) {
                    foundURI = true;
                    uri = buf.toString();
                    buf.setLength(0);
                    inCurly = false;
                }
                else {
                    buf.append(c);
                }
            }
            if (buf.length() > 0) {
                localName = buf.toString();
                if (!foundURI) {
                    uri = "";
                }
                this.addCDATAElement(uri, localName);
            }
        }
    }
    
    private void addCDATAElement(String uri, String localName) {
        if (this.m_CdataElems == null) {
            this.m_CdataElems = new HashMap();
        }
        HashMap<String, String> h = (HashMap)this.m_CdataElems.get(localName);
        if (h == null) {
            h = new HashMap();
            this.m_CdataElems.put(localName, h);
        }
        h.put(uri, uri);
    }
    
    public boolean documentIsEmpty() {
        return this.m_docIsEmpty && this.m_elemContext.m_currentElemDepth == 0;
    }
    
    protected boolean isCdataSection() {
        boolean b = false;
        if (null != this.m_StringOfCDATASections) {
            if (this.m_elemContext.m_elementLocalName == null) {
                String localName = getLocalName(this.m_elemContext.m_elementName);
                this.m_elemContext.m_elementLocalName = localName;
            }
            if (this.m_elemContext.m_elementURI == null) {
                this.m_elemContext.m_elementURI = this.getElementURI();
            }
            else if (this.m_elemContext.m_elementURI.length() == 0) {
                if (this.m_elemContext.m_elementName == null) {
                    this.m_elemContext.m_elementName = this.m_elemContext.m_elementLocalName;
                }
                else if (this.m_elemContext.m_elementLocalName.length() < this.m_elemContext.m_elementName.length()) {
                    this.m_elemContext.m_elementURI = this.getElementURI();
                }
            }
            HashMap<String, String> h = null;
            if (this.m_CdataElems != null) {
                h = (HashMap)this.m_CdataElems.get(this.m_elemContext.m_elementLocalName);
            }
            if (h != null) {
                Object obj = h.get(this.m_elemContext.m_elementURI);
                if (obj != null) {
                    b = true;
                }
            }
        }
        return b;
    }
    
    private String getElementURI() {
        String uri = null;
        String prefix = getPrefixPart(this.m_elemContext.m_elementName);
        if (prefix == null) {
            uri = this.m_prefixMap.lookupNamespace("");
        }
        else {
            uri = this.m_prefixMap.lookupNamespace(prefix);
        }
        if (uri == null) {
            uri = "";
            return uri;
        }
        return uri;
    }
    
    @Override
    public String getOutputProperty(String name) {
        String val = this.getOutputPropertyNonDefault(name);
        if (val == null) {
            val = this.getOutputPropertyDefault(name);
            return val;
        }
        return val;
    }
    
    public String getOutputPropertyNonDefault(String name) {
        return this.getProp(name, false);
    }
    
    @Override
    public Object asDOM3Serializer() throws IOException {
        return new DOM3SerializerImpl(this);
    }
    
    @Override
    public String getOutputPropertyDefault(String name) {
        return this.getProp(name, true);
    }
    
    @Override
    public void setOutputProperty(String name, String val) {
        this.setProp(name, val, false);
    }
    
    @Override
    public void setOutputPropertyDefault(String name, String val) {
        this.setProp(name, val, true);
    }
    
    Set<String> getOutputPropDefaultKeys() {
        return this.m_OutputPropsDefault.keySet();
    }
    
    Set<String> getOutputPropKeys() {
        return this.m_OutputProps.keySet();
    }
    
    private String getProp(String name, boolean defaultVal) {
        if (this.m_OutputProps == null) {
            this.m_OutputProps = new HashMap();
            this.m_OutputPropsDefault = new HashMap();
        }
        if (defaultVal) {
            String val = (String)this.m_OutputPropsDefault.get(name);
            return val;
        }
        String val = (String)this.m_OutputProps.get(name);
        return val;
    }
    
    void setProp(String name, String val, boolean defaultVal) {
        if (this.m_OutputProps == null) {
            this.m_OutputProps = new HashMap();
            this.m_OutputPropsDefault = new HashMap();
        }
        if (defaultVal) {
            this.m_OutputPropsDefault.put(name, val);
        }
        else if ("cdata-section-elements".equals(name) && val != null) {
            this.initCdataElems(val);
            String oldVal = (String)this.m_OutputProps.get(name);
            String newVal;
            if (oldVal == null) {
                newVal = oldVal + 32 + val;
            }
            else {
                newVal = val;
            }
            this.m_OutputProps.put(name, newVal);
        }
        else {
            this.m_OutputProps.put(name, val);
        }
    }
    
    static char getFirstCharLocName(String name) {
        int i = name.indexOf(125);
        char first;
        if (i < 0) {
            first = name.charAt(0);
        }
        else {
            first = name.charAt(i + 1);
        }
        return first;
    }
}
